"""Combine all the components of the dashboard into a single app."""
import panel as pn
from .data import DataManager
from .widgets import Widgets
from .layout import Layout
from .callbacks import Callbacks

class Dashboard:
    def __init__(self, title: str):
        self.data_manager = DataManager()
        self.widgets = Widgets()
        self.layout = Layout(title, self.widgets)
        self.callbacks = Callbacks(
            data_manager=self.data_manager,
            layout=self.layout,
            widgets=self.widgets
        )

    def serve(self):
        """
        Serve the dashboard.
        """
        pn.serve(self.layout.servable())



"""Layout widgets for the dashboard."""
from typing import Protocol
import pandas as pd
import panel as pn
from panel.template import BootstrapTemplate
from .widgets import Widgets

class TableBuilder(Protocol):
    """
    Build a metrics table from the given data.
    
    Parameters
    ----------
    data: pd.DataFrame
        Data to display in the metrics table.
    
    Returns
    -------
    pn.widgets.Tabulator
        A Tabulator widget displaying the metrics table.
    """
    @staticmethod
    def __call__(data: pd.DataFrame) -> pn.widgets.Tabulator:
        ...

class Layout:
    """
    Layout for the dashboard.
    
    Attributes
    ----------
    widgets: Widgets
        Instance of Widgets to create the layout.
    tabs: pn.Tabs
        Dashboard tabs.
    template: pn.template
        Servable dashboard with widgets laid out.
    """
    def __init__(self, title: str, widgets: Widgets):
        """
        Initialize the layout of the dashboard.
        
        Parameters
        ----------
        title: str
            Dashboard title.
        widgets: Widgets
            Instance of Widgets to create the layout.
        """
        self.widgets = widgets
        self.tabs = pn.Tabs()
        self.add_tab(
            "File Selector",
            pn.Column(self.widgets.file_selector, self.widgets.load_data_button)
            )
        self.add_tab(
            "Feature Selector",
            pn.Row(
                pn.Column(
                    self.widgets.left_feature_selector,
                    self.widgets.right_feature_selector,
                    self.widgets.description_pane
                    ),
                self.widgets.map_selector
            )
        )
        self.add_tab(
            "Metrics Plots",
            pn.Row(
                pn.Column(
                    self.widgets.description_pane,
                    self.widgets.selected_metric
                ),
                self.widgets.metrics_pane
            )
        )
        self.add_tab(
            "Pairs Plots",
            self.widgets.pairs_pane
        )
        self.add_tab(
            "Metrics Table",
            self.widgets.build_table(
                pd.DataFrame({"message": ["no data loaded"]})
        ))
        self.add_tab(
            "Pairs Table",
            self.widgets.build_table(
                pd.DataFrame({"message": ["no data loaded"]})
        ))
        self.template = BootstrapTemplate(title=title)
        self.template.main.append(self.tabs)
    
    def add_tab(self, name: str, content: pn.pane) -> None:
        """
        Add a tab to the tabs panel.
        
        Parameters
        ----------
        name: str
            Name of the tab.
        content: pn.pane
            Content of the tab.
        """
        self.tabs.append((name, content))
    
    def servable(self) -> BootstrapTemplate:
        """
        Serve the layout.
        """
        return self.template.servable()

    def update_metrics_table(
            self,
            data: pd.DataFrame,
            table_builder: TableBuilder,
            left_feature_name: str | None = None,
            metric_name: str | None = None
            ) -> None:
        """
        Update metrics table with new data.
        
        Parameters
        ----------
        data: pd.DataFrame
            Data to display in the metrics table.
        """
        df = data
        if left_feature_name is not None:
            df = df[df["LEFT FEATURE NAME"] == left_feature_name]
        if metric_name is not None:
            df = df[df["METRIC NAME"] == metric_name]

        self.tabs[4] = (
            "Metrics Table",
            table_builder(df)
            )

    def update_pairs_table(
            self,
            data: pd.DataFrame,
            table_builder: TableBuilder,
            feature_name: str | None = None
            ) -> None:
        """
        Update pairs table with new data.
        
        Parameters
        ----------
        data: pd.DataFrame
            Data to display in the pairs table.
        """
        if feature_name is not None:
            df = data[data["FEATURE NAME"] == feature_name]
        else:
            df = data

        self.tabs[5] = (
            "Pairs Table",
            table_builder(df)
            )



"""Methods to load and process WRES output."""
from typing import Type, Iterable
import pandas as pd
import geopandas as gpd

METRICS_COLUMNS: dict[str, Type] = {
    "LEFT VARIABLE NAME": "category",
    "RIGHT VARIABLE NAME": "category",
    "BASELINE VARIABLE NAME": str,
    "COVARIATE FILTERS": str,
    "POOL NUMBER": int,
    "EVALUATION SUBJECT": "category",
    "FEATURE GROUP NAME": "category",
    "LEFT FEATURE NAME": "category",
    "LEFT FEATURE WKT": "category",
    "LEFT FEATURE SRID": "category",
    "LEFT FEATURE DESCRIPTION": "category",
    "RIGHT FEATURE NAME": "category",
    "RIGHT FEATURE WKT": "category",
    "RIGHT FEATURE SRID": "category",
    "RIGHT FEATURE DESCRIPTION": "category",
    "BASELINE FEATURE NAME": "category",
    "BASELINE FEATURE WKT": "category",
    "BASELINE FEATURE SRID": "category",
    "BASELINE FEATURE DESCRIPTION": "category",
    "EARLIEST ISSUED TIME EXCLUSIVE": "datetime",
    "LATEST ISSUED TIME INCLUSIVE": "datetime",
    "EARLIEST VALID TIME EXCLUSIVE": "datetime",
    "LATEST VALID TIME INCLUSIVE": "datetime",
    "EARLIEST LEAD DURATION EXCLUSIVE": "category",
    "LATEST LEAD DURATION INCLUSIVE": "category",
    "TIME SCALE DURATION": "category",
    "TIME SCALE FUNCTION": "category",
    "TIME SCALE START MONTH-DAY INCLUSIVE": str,
    "TIME SCALE END MONTH-DAY INCLUSIVE": str,
    "EVENT THRESHOLD NAME": str,
    "EVENT THRESHOLD LOWER VALUE": "numeric",
    "EVENT THRESHOLD UPPER VALUE": str,
    "EVENT THRESHOLD UNITS": str,
    "EVENT THRESHOLD LOWER PROBABILITY": str,
    "EVENT THRESHOLD UPPER PROBABILITY": str,
    "EVENT THRESHOLD SIDE": "category",
    "EVENT THRESHOLD OPERATOR": "category",
    "DECISION THRESHOLD NAME": str,
    "DECISION THRESHOLD LOWER VALUE": str,
    "DECISION THRESHOLD UPPER VALUE": str,
    "DECISION THRESHOLD UNITS": str,
    "DECISION THRESHOLD LOWER PROBABILITY": str,
    "DECISION THRESHOLD UPPER PROBABILITY": str,
    "DECISION THRESHOLD SIDE": str,
    "DECISION THRESHOLD OPERATOR": str,
    "METRIC NAME": "category",
    "METRIC COMPONENT NAME": "category",
    "METRIC COMPONENT QUALIFIER": str,
    "METRIC COMPONENT UNITS": "category",
    "METRIC COMPONENT MINIMUM": "numeric",
    "METRIC COMPONENT MAXIMUM": "numeric",
    "METRIC COMPONENT OPTIMUM": "numeric",
    "STATISTIC GROUP NUMBER": int,
    "SUMMARY STATISTIC NAME": "category",
    "SUMMARY STATISTIC COMPONENT NAME": "category",
    "SUMMARY STATISTIC UNITS": "category",
    "SUMMARY STATISTIC DIMENSION": "category",
    "SUMMARY STATISTIC QUANTILE": str,
    "SAMPLE QUANTILE": "numeric",
    "STATISTIC": "numeric",
    "EVALUATION PERIOD": "category",
    "LEAD TIME": "category"
}
"""WRES metrics columns and data types."""

PAIRS_COLUMNS: dict[str, Type] = {
    "FEATURE NAME": "category",
    "FEATURE GROUP NAME": "category",
    "VARIABLE NAME": "category",
    "EARLIEST ISSUE TIME": "datetime",
    "LATEST ISSUE TIME": "datetime",
    "EARLIEST VALID TIME": "datetime",
    "LATEST VALID TIME": "datetime",
    "EARLIEST LEAD TIME [MAXIMUM OVER PAST PT24H]": "category",
    "LATEST LEAD TIME [MAXIMUM OVER PAST PT24H]": "category",
    "REFERENCE TIME": "datetime",
    "VALID TIME": "datetime",
    "LEAD DURATION [MAXIMUM OVER PAST PT24H]": "category",
    "OBSERVED IN ft3/s": "numeric",
    "PREDICTED IN ft3/s": "numeric"
}
"""WRES pairs columns and data types."""

def sort_filepaths(filepaths: Iterable[str]) -> tuple[list[str], list[str]]:
    """
    Sort filepaths into metrics and pairs file list.
    
    Parameters
    ----------
    filepaths: Iterable[str], required
        Paths to files.
    
    Returns
    -------
    metrics_filepaths, pairs_filepaths: tuple[list[str], list[str]]
    """
    metrics_filepaths = []
    pairs_filepaths = []
    
    for filepath in filepaths:
        df = pd.read_csv(filepath, nrows=1)
        if "STATISTIC" in df:
            metrics_filepaths.append(filepath)
        if "PREDICTED IN ft3/s" in df:
            pairs_filepaths.append(filepath)
    return metrics_filepaths, pairs_filepaths

def load_dataframes(
        filepaths: Iterable[str],
        type_mapping: dict[str, Type] = METRICS_COLUMNS
        ) -> pd.DataFrame:
    """
    Load CSV output and return optimized dataframe.
    
    Parameters
    ----------
    filepaths: Iterable[str], required
        Paths to files.
    type_mapping: dict[str, Type], optional
        Mapping from column label to data type.
    
    Returns
    -------
    pandas.DataFrame
    """
    data = pd.concat(
        [pd.read_csv(fp, dtype=str) for fp in filepaths],
        ignore_index=True
        )

    drop = []
    for c in data:
        if data[c].isna().all():
            drop.append(c)
            continue

        if type_mapping[c] == "datetime":
            data[c] = pd.to_datetime(data[c])
            continue

        if type_mapping[c] == "numeric":
            data[c] = pd.to_numeric(data[c], errors="coerce")
            continue

        if type_mapping[c] == "geometry":
            data[c] = gpd.GeoSeries.from_wkt(data[c])
            continue

        data[c] = data[c].astype(type_mapping[c])
    
    return data.drop(drop, axis=1)

def load_metrics_dataframes(
        filepaths: Iterable[str],
        type_mapping: dict[str, Type] = METRICS_COLUMNS
        ) -> pd.DataFrame:
    """
    Load metrics CSV output and return optimized dataframe.
    
    Parameters
    ----------
    filepaths: Iterable[str], required
        Paths to files.
    type_mapping: dict[str, Type], optional
        Mapping from column label to data type.
    
    Returns
    -------
    pandas.DataFrame
    """
    if len(filepaths) == 0:
        return pd.DataFrame({"message": ["no data loaded"]})

    data = load_dataframes(filepaths, type_mapping)

    data["EVALUATION PERIOD"] = (
        data["LATEST ISSUED TIME INCLUSIVE"] -
        data["EARLIEST ISSUED TIME EXCLUSIVE"]
    ).astype(str).astype("category")
    earliest = data["EARLIEST LEAD DURATION EXCLUSIVE"].str.extract("(\\d+)")
    latest = data["LATEST LEAD DURATION INCLUSIVE"].str.extract("(\\d+)")
    data["LEAD HOURS MIN"] = earliest.astype(int)
    data["LEAD HOURS MAX"] = latest.astype(int)

    return data

def load_pairs_dataframes(
        filepaths: Iterable[str],
        type_mapping: dict[str, Type] = PAIRS_COLUMNS
        ) -> pd.DataFrame:
    """
    Load pairs CSV output and return optimized dataframe.
    
    Parameters
    ----------
    filepaths: Iterable[str], required
        Paths to files.
    type_mapping: dict[str, Type], optional
        Mapping from column label to data type.
    
    Returns
    -------
    pandas.DataFrame
    """
    if len(filepaths) == 0:
        return pd.DataFrame({"message": ["no data loaded"]})
    data = load_dataframes(filepaths, type_mapping)
    return data

class DataManager:
    """
    Handle the data loading and processing.

    Attributes
    ----------
    data: pd.DataFrame
        Data loaded from the CSV files.
    pairs: pd.DataFrame
        Pairs data loaded from the CSV files.
    feature_mapping: pd.DataFrame
        Mapping of features to their descriptions and geometries.
    """
    def __init__(self):
        self.data: pd.DataFrame = None
        self.pairs: pd.DataFrame = None
        self.feature_mapping: pd.DataFrame = None
    
    def load_data(self, filepaths: list[str]):
        if len(filepaths) == 0:
            self.data = pd.DataFrame({"message": ["no data loaded"]})
            self.feature_mapping = pd.DataFrame(
                {"message": ["no data loaded"]})
            self.pairs = pd.DataFrame({"message": ["no data loaded"]})
        else:
            try:
                metrics_files, pairs_files = sort_filepaths(filepaths)
                self.data = load_metrics_dataframes(metrics_files)
                self.pairs = load_pairs_dataframes(pairs_files)
                self.feature_mapping = self.data[[
                    "LEFT FEATURE NAME",
                    "LEFT FEATURE DESCRIPTION",
                    "RIGHT FEATURE NAME",
                    "LEFT FEATURE WKT"
                    ]].drop_duplicates().astype(str)
                self.feature_mapping["geometry"] = gpd.GeoSeries.from_wkt(
                    self.feature_mapping["LEFT FEATURE WKT"])
                self.feature_mapping = gpd.GeoDataFrame(self.feature_mapping)
            except pd.errors.ParserError:
                self.data = pd.DataFrame({"message": ["parsing error"]})
                self.pairs = pd.DataFrame({"message": ["parsing error"]})
            except KeyError:
                self.data = pd.DataFrame({"message": ["column error"]})
                self.pairs = pd.DataFrame({"message": ["column error"]})



"""Launch the dashboard from a CLI."""
import click
from .dashboard import Dashboard

@click.command()
def run() -> None:
    """
    Visualize and explore metrics output from WRES CSV2 formatted output.

    Run "wres-explorer" from the command-line, ctrl+c to stop the server.:
    """
    # Start interface
    Dashboard("WRES CSV Explorer").serve()

if __name__ == "__main__":
    run()



"""Plots for the dashboard."""
import pandas as pd
import geopandas as gpd
import plotly.graph_objects as go
import colorcet as cc

def generate_map(geodata: gpd.GeoDataFrame) -> go.Figure:
    """
    Generate a map of points.

    Parameters
    ----------
    geodata: geopandas.GeoDataFrame
        One-to-One feature mapping with WRES CSV2-compatible column names.
        Required columns include: ['geometry', 'LEFT FEATURE NAME', 
        'LEFT FEATURE DESCRIPTION', 'RIGHT FEATURE NAME', 'LONGITUDE',
        'LATITUDE']
    """
    if "geometry" not in geodata:
        return go.Figure()
    
    # Build map
    fig = go.Figure(go.Scattermap(
        showlegend=False,
        name="",
        lat=geodata["geometry"].y,
        lon=geodata["geometry"].x,
        mode="markers",
        marker=dict(
            size=15,
            color="cyan"
            ),
        selected=dict(
            marker=dict(
                color="magenta"
            )
        ),
        customdata=geodata[[
            "LEFT FEATURE NAME",
            "LEFT FEATURE DESCRIPTION",
            "RIGHT FEATURE NAME"
            ]],
        hovertemplate=
        "LEFT FEATURE DESCRIPTION: %{customdata[1]}<br>"
        "LEFT FEATURE NAME: %{customdata[0]}<br>"
        "RIGHT FEATURE NAME: %{customdata[2]}<br>"
        "LONGITUDE: %{lon}<br>"
        "LATITUDE: %{lat}<br>"
    ))
    fig.update_layout(
        showlegend=False,
        height=720,
        width=1280,
        margin=dict(l=0, r=0, t=0, b=0),
        map=dict(
            style="satellite-streets",
            center={
                "lat": geodata["geometry"].y.mean(),
                "lon": geodata["geometry"].x.mean()
                },
            zoom=2
        ),
        clickmode="event+select",
        modebar=dict(
            remove=["lasso", "select"]
        ),
        dragmode="zoom"
    )
    return fig

def generate_metrics_plot(
        data: pd.DataFrame,
        left_feature_name: str,
        selected_metric: str
    ) -> go.Figure:
    """
    Generate a metrics plot.

    Parameters
    ----------
    data: pd.DataFrame
        Data containing metrics information.
    left_feature_name: str
        Name of the left feature to filter the data.
    selected_metric: str
        Name of the metric to plot.

    Returns
    -------
    go.Figure
        Plotly figure object containing the metrics plot.
    """
    if "LEFT FEATURE NAME" not in data:
        return go.Figure()
    if "METRIC NAME" not in data:
        return go.Figure()

    # Subset data for the selected feature and metric
    df = data[data["LEFT FEATURE NAME"] == left_feature_name]
    df = df[df["METRIC NAME"] == selected_metric]

    fig = go.Figure()
    
    for period, d in df.groupby("EVALUATION PERIOD", observed=True):
        xmin = d[d["SAMPLE QUANTILE"].isna()]["LEAD HOURS MIN"].values
        xmax = d[d["SAMPLE QUANTILE"].isna()]["LEAD HOURS MAX"].values
        nom_y = d[d["SAMPLE QUANTILE"].isna()]["STATISTIC"].values
        upper = d[d["SAMPLE QUANTILE"] == 0.975]["STATISTIC"].values
        lower = d[d["SAMPLE QUANTILE"] == 0.025]["STATISTIC"].values
        
        if len(nom_y) == len(upper) == len(lower):
            error_y = dict(
                type="data",
                array=upper - nom_y,
                arrayminus=nom_y - lower
            )
        else:
            error_y = None
        
        fig.add_trace(go.Bar(
            name=period,
            x=xmin, y=nom_y,
            error_y=error_y,
            legendgroup="bar_plots",
            legendgrouptitle_text="Evaluation Period"
        ))
    
    # Determine ticks
    xmin = sorted(data["LEAD HOURS MIN"].unique().tolist())
    xmax = sorted(data["LEAD HOURS MAX"].unique().tolist())
    xticks = [f"{e}-{l}" for e, l in zip(xmin, xmax)]
    
    fig.update_xaxes(title="LEAD HOURS")
    fig.update_yaxes(title=selected_metric)
    fig.update_layout(
        height=720,
        width=1280,
        margin=dict(l=0, r=0, t=0, b=0),
        xaxis=dict(
            tickmode="array",
            tickvals=sorted(xmin),
            ticktext=xticks
        )
    )
    return fig

def generate_pairs_plot(
        data: pd.DataFrame,
        feature_name: str
    ) -> go.Figure:
    """
    Generate a pairs plot.

    Parameters
    ----------
    data: pd.DataFrame
        Data containing pairs information.
    feature_name: str
        Name of the feature to filter the data.

    Returns
    -------
    go.Figure
        Plotly figure object containing the pairs plot.
    """
    if "FEATURE NAME" not in data:
        return go.Figure()

    # Subset data for the selected feature and metric
    df = data[data["FEATURE NAME"] == feature_name]

    # Parse out observations
    obs = df[["VALID TIME", "OBSERVED IN ft3/s"]].drop_duplicates()

    # Parse out predicted
    pred = df[[
        "REFERENCE TIME",
        "VALID TIME",
        "PREDICTED IN ft3/s"
    ]].drop_duplicates().sort_values(by=["REFERENCE TIME", "VALID TIME"])

    fig = go.Figure()

    fig.update_xaxes(title="VALID TIME")
    fig.update_yaxes(title="STREAMFLOW (ft³/s)")
    fig.update_layout(
        height=720,
        width=1280,
        margin=dict(l=0, r=0, t=0, b=0)
    )

    fig.add_trace(go.Scatter(
        x=obs["VALID TIME"],
        y=obs["OBSERVED IN ft3/s"],
        mode="markers",
        name="Observed",
        marker=dict(color="cyan", line=dict(color="black", width=1)),
        hovertemplate=
        "VALID TIME: %{x}<br>"
        "OBSERVED: %{y}<br>",
        legendgroup="observed"
    ))

    idx = 0
    for rt, p in pred.groupby("REFERENCE TIME"):
        time_str = rt.strftime("%Y-%m-%d %H")
        fig.add_trace(go.Scatter(
            x=p["VALID TIME"],
            y=p["PREDICTED IN ft3/s"],
            mode="lines",
            name=time_str,
            line=dict(color=cc.CET_L8[idx]),
            hovertemplate=
            f"REFERENCE TIME: {time_str}<br>"
            "VALID TIME: %{x}<br>"
            "PREDICTED: %{y}<br>",
            legendgroup="predicted",
            legendgrouptitle_text="Predicted"
        ))
        idx += 1
        if idx == len(cc.CET_L8):
            idx = 0
    
    return fig



"""Callbacks for dashboard."""
import panel as pn
from .data import DataManager
from .layout import Layout
from .widgets import Widgets
from .plots import generate_map, generate_metrics_plot, generate_pairs_plot

class Callbacks:
    """Class to handle callbacks for the dashboard.
    
    Attributes
    ----------
    data_manager : DataManager
        Instance of DataManager to handle data loading and processing.
    layout : Layout
        Instance of Layout to manage the layout of the dashboard.
    widgets : Widgets
        Instance of Widgets to manage the widgets in the dashboard.
    """
    def __init__(
            self,
            data_manager: DataManager,
            layout: Layout,
            widgets: Widgets
        ) -> None:
        self.data_manager = data_manager
        self.layout = layout
        self.widgets = widgets
        self.feature_descriptions = []

        # Callback for loading data
        def load_data(event):
            if not event:
                return
            self.data_manager.load_data(self.widgets.file_selector.value)
            self.layout.update_metrics_table(
                self.data_manager.data,
                self.widgets.build_table
            )
            self.layout.update_pairs_table(
                self.data_manager.pairs,
                self.widgets.build_table
            )
            self.widgets.map_selector.object = generate_map(
                self.data_manager.feature_mapping
            )
            self.update_feature_selectors()
            self.update_metric_selector()
        pn.bind(load_data, self.widgets.load_data_button, watch=True)

        # Link feature selectors
        def update_left(right_value):
            if not right_value:
                return
            idx = self.widgets.right_feature_selector.options.index(right_value)
            self.widgets.left_feature_selector.value = (
                self.widgets.left_feature_selector.options[idx]
                )
            self.widgets.description_pane.object = (
                    "LEFT FEATURE DESCRIPTION<br>" +
                    self.feature_descriptions[idx]
                )
        def update_right(left_value):
            if not left_value:
                return
            idx = self.widgets.left_feature_selector.options.index(left_value)
            self.widgets.right_feature_selector.value = (
                self.widgets.right_feature_selector.options[idx]
                )
            self.widgets.description_pane.object = (
                    "LEFT FEATURE DESCRIPTION<br>" +
                    self.feature_descriptions[idx]
                )
        def update_from_map(event):
            if not event:
                return
            try:
                point = event["points"][0]
                self.widgets.left_feature_selector.value = point["customdata"][0]
                self.widgets.right_feature_selector.value = point["customdata"][2]
                self.widgets.description_pane.object = (
                    "LEFT FEATURE DESCRIPTION<br>" +
                    point["customdata"][1]
                )
            except Exception as ex:
                self.widgets.description_pane.object = (
                    f"Could not determine site selection: {ex}"
                )
        pn.bind(update_from_map, self.widgets.map_selector.param.click_data, watch=True)
        pn.bind(update_left, right_value=self.widgets.right_feature_selector,
                watch=True)
        pn.bind(update_right, left_value=self.widgets.left_feature_selector,
                watch=True)
        
        # Link metric selector to metrics pane
        def update_metrics_plot(event):
            fig = generate_metrics_plot(
                self.data_manager.data,
                self.widgets.left_feature_selector.value,
                self.widgets.selected_metric.value
            )
            self.widgets.metrics_pane.object = fig
        pn.bind(
            update_metrics_plot,
            self.widgets.selected_metric,
            watch=True
        )
        pn.bind(
            update_metrics_plot,
            self.widgets.left_feature_selector,
            watch=True
        )
    
        # Filter tables by feature selection
        def filter_by_feature(left: str):
            if left is None:
                return
            if "LEFT FEATURE NAME" in self.data_manager.data:
                self.layout.update_metrics_table(
                    self.data_manager.data,
                    self.widgets.build_table,
                    left_feature_name=left
                )
            if "FEATURE NAME" in self.data_manager.pairs:
                self.layout.update_pairs_table(
                    self.data_manager.pairs,
                    self.widgets.build_table,
                    feature_name=left
                )
        pn.bind(
            filter_by_feature,
            self.widgets.left_feature_selector,
            watch=True
        )
    
        # Filter metrics table by selected metric
        def filter_by_metric(metric: str):
            if metric is None:
                return
            if "LEFT FEATURE NAME" in self.data_manager.data:
                self.layout.update_metrics_table(
                    self.data_manager.data,
                    self.widgets.build_table,
                    left_feature_name=self.widgets.left_feature_selector.value,
                    metric_name=metric
                )
        pn.bind(
            filter_by_metric,
            self.widgets.selected_metric,
            watch=True
        )
        
        # Link feature selector to pairs pane
        def update_pairs_plot(event):
            fig = generate_pairs_plot(
                self.data_manager.pairs,
                self.widgets.left_feature_selector.value
            )
            self.widgets.pairs_pane.object = fig
        pn.bind(
            update_pairs_plot,
            self.widgets.left_feature_selector,
            watch=True
        )
    
    def update_feature_selectors(self) -> None:
        if "LEFT FEATURE NAME" not in self.data_manager.feature_mapping:
            self.widgets.left_feature_selector.options = []
            self.widgets.right_feature_selector.options = []
            self.feature_descriptions = []
            self.widgets.left_feature_selector.value = None
            self.widgets.right_feature_selector.value = None
            self.widgets.description_pane.object = (
                "LEFT FEATURE DESCRIPTION<br>"
                "No data loaded"
            )
            return
        self.widgets.left_feature_selector.options = (
            self.data_manager.feature_mapping[
                "LEFT FEATURE NAME"].tolist())
        self.widgets.right_feature_selector.options = (
            self.data_manager.feature_mapping[
                "RIGHT FEATURE NAME"].tolist())
        self.feature_descriptions = (
            self.data_manager.feature_mapping[
                "LEFT FEATURE DESCRIPTION"].tolist())
    
    def update_metric_selector(self) -> None:
        if "METRIC NAME" not in self.data_manager.data:
            self.widgets.selected_metric.options = []
            return
        self.widgets.selected_metric.options = (
            self.data_manager.data["METRIC NAME"].unique().tolist())



"""Widgets for the dashboard."""
import panel as pn
import pandas as pd

pn.extension("tabulator")

class Widgets:
    """
    Widgets for the dashboard.
    
    Attributes
    ----------
    file_selector: pn.widgets.FileSelector
        File selector widget for selecting CSV files.
    load_data_button: pn.widgets.Button
        Button to load/reload data.
    left_feature_selector: pn.widgets.AutocompleteInput
        Autocomplete input for selecting left feature.
    right_feature_selector: pn.widgets.AutocompleteInput
        Autocomplete input for selecting right feature.
    map_selector: pn.pane.Plotly
        Pane for displaying the map of features.
    description_pane: pn.pane.Markdown
        Pane for displaying feature descriptions.
    selected_metric: pn.widgets.Select
        Select widget for selecting metrics.
    metrics_pane: pn.pane.Plotly
        Pane for displaying metrics plots.
    pairs_pane: pn.pane.Plotly
        Pane for displaying pairs plots.
    
    Methods
    -------
    build_metrics_table
    """
    def __init__(self):
        self.file_selector = pn.widgets.FileSelector(
            directory="./",
            file_pattern="*.csv.gz",
            only_files=True,
            value=[]
        )
        self.load_data_button = pn.widgets.Button(
            name="Load/Reload Data",
            button_type="primary"
        )
        self.left_feature_selector = pn.widgets.AutocompleteInput(
            name="LEFT FEATURE NAME",
            options=[],
            search_strategy="includes",
            placeholder=f"Select LEFT FEATURE NAME"
        )
        self.right_feature_selector = pn.widgets.AutocompleteInput(
            name="RIGHT FEATURE NAME",
            options=[],
            search_strategy="includes",
            placeholder=f"Select RIGHT FEATURE NAME"
        )
        self.description_pane = pn.pane.Markdown(
            "LEFT FEATURE DESCRIPTION<br>"
            "No data loaded"
        )
        self.map_selector = pn.pane.Plotly()
        self.selected_metric = pn.widgets.Select(
            name="Select Metric",
            options=[]
        )
        self.metrics_pane = pn.pane.Plotly()
        self.pairs_pane = pn.pane.Plotly()
    
    @staticmethod
    def build_table(data: pd.DataFrame) -> pn.widgets.Tabulator:
        """
        Build a table with the provided data.
        
        Parameters
        ----------
        data: pd.DataFrame
            Data to display in the table.
        """
        return pn.widgets.Tabulator(
            data,
            show_index=False,
            disabled=True,
            width=1280,
            height=720
        )
